/*
 * dlock.c - Solaris lslk lock functions
 *
 * Vic Abell
 * Purdue University Computing Center
 */


/*
 * Copyright 1996 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell.
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */
#ifndef lint
static char copyright[] =
"@(#) Copyright 1996 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: dlock.c,v 1.14 2001/01/08 18:21:03 abe Exp $";
#endif


#include "lslk.h"
#include "kernelbase.h"

/*
 * Local function prototypes
 */

_PROTOTYPE(static int readsysids,(void));
_PROTOTYPE(static void savelock,(struct inode *i, struct lock_descriptor *ld));


/*
 * gather_lock_info() -- gather lock information
 */

void
gather_lock_info()
{
	int hx;
	struct inode i, *ip;
	union ihead ih, *ihp;
	struct lock_descriptor ld;
	KA_T ldp;
	vnode_t *va;
/*
 * Loop through the inode hash buckets.
 */
	for (hx = 0, ihp = Ihdp; hx < Ihsz; hx++, ihp++) {
	    if (kread((KA_T)ihp, (char *)&ih, sizeof(ih))) {
		if (Owarn)
		    (void) fprintf(stderr, "%s: can't read ihead\n", Pn);
		return;
	    }
	/*
	 * Loop through the inodes of a hash bucket.
	 */
	    for (ip = ih.ih_chain[0];
	         ip && ip != (struct inode *)ihp;
	         ip = i.i_forw)
	    {
		if (kread((KA_T)ip, (char *)&i, sizeof(i))) {
		    if (Owarn)
			(void) fprintf(stderr,
			    "%s: premature exit from inode chain\n", Pn);
		    return;
		}
		if (!i.i_vnode.v_filocks || !i.i_vnode.v_count)
		    continue;
	    /*
	     * If the vnode, contained in the inode, has a non-NULL v_filocks
	     * pointer and a non-zero reference count, read and list the
	     * associated lock_descriptor structure chain.
	     *
	     * NB: this *seems* to work, mostly because we're looking
	     *     at inodes.  Under Solaris 2.3 (after patch 101318-45)
	     *     and all patch levels of 2.4, when the vnode refers to
	     *     an NFS file (an rnode), the v_filocks pointer addresses
	     *     a filock structure.
	     */
		va = (vnode_t *)((char *)ip + offsetof(struct inode, i_vnode));
		ldp = (KA_T)i.i_vnode.v_filocks;
		do {
		    if (kread((KA_T)ldp, (char *)&ld, sizeof(ld)))
			ldp = (KA_T)i.i_vnode.v_filocks;
		    else {
			ldp = (KA_T)ld.LOCK_NEXT;
			if ((ld.LOCK_FLAGS & ACTIVE_LOCK)
			&&  va == ld.LOCK_VNODE) {

	    		/*
	     		 * If the lock_descriptor references an active lock,
			 * save its information.
	     		 */
			    (void) savelock(&i, &ld);
			    if (is_lock_sel()) {
				NLockU++;
				Lp = (struct llock_info *)NULL;
			    }
			}
		    }
		} while (ldp && ldp != (KA_T)i.i_vnode.v_filocks);
	    }
	}
}


/*
 * get_cmdnm() -- get command name
 */

char *
get_cmdnm(lp)
	struct llock_info *lp;		/* local lock structure */
{
	struct proc *p;

/*
 * Get the proc structure for the lock info structure and return the
 * command name of its user structure.
 */
	if (lp->pid && (p = kvm_getproc(Kd, lp->pid)))
	    return(p->p_user.u_comm);
	return("(unknown)");
}


/*
 * get_kernel_access() -- get access to kernel information
 */

void
get_kernel_access()
{
	int err, i;

#if	solaris>=70000
	char *cp, isa[1024];
	short kbits = 32;

# if	defined(_LP64)
	short xkbits = 64;
# else	/* !defined(_LP64) */
	short xkbits = 32;
# endif	/* defined(_LP64) */

/*
 * Compare the Solaris 7 or higher lslk compilation bit size with the kernel
 * bit size.  Quit on a mismatch.
 */
	if (sysinfo(SI_ISALIST, isa, (long)sizeof(isa)) < 0) {
	    (void) fprintf(stderr, "%s: can't get ISA list: %s\n",
		Pn, strerror(errno));
	    Exit(1);
	}
	for (cp = isa; *cp;) {
	    if (strncmp(cp, "sparcv9", strlen("sparcv9")) == 0) {
		kbits = 64;
		break;
	    }
	    if (!(cp = strchr(cp, ' ')))
		break;
	    cp++;
	}
	if (kbits != xkbits) {
	    (void) fprintf(stderr,
		"%s: FATAL: lsof was compiled for a %d bit kernel,\n",
		Pn, (int)xkbits);
	    (void) fprintf(stderr,
		"      but this machine has booted a %d bit kernel.\n",
		(int)kbits);
	    Exit(1);
	}
#endif	/* solaris>=70000 */

/*
 * Convert kernel variable names to addresses and check the result.
 */
	if (Nmlst && !is_readable(Nmlst, 1))
	    Exit(1);
	if (nlist(Nmlst ? Nmlst : N_UNIX, Nl) < 0) {
	    (void) fprintf(stderr, "%s: nlist(%s) failed\n", Pn,
		Nmlst ? Nmlst : N_UNIX);
	    Exit(1);
	}
	for (err = i = 0; Nl[i].n_name[0]; i++) {
	  if (!Nl[i].n_value) {

#if	solaris>=20500
	    if (i == X_SYSIDS) {
	      if (Owarn) {
		(void) fprintf(stderr,
		  "%s: no system ID to IP address table found;\n", Pn);
		(void) fprintf(stderr,
		  "      Remote sources will be identified by system ID.\n");
	      }
	      continue;
	    }
#endif	/* solaris>=20500 */

	    (void) fprintf(stderr, "%s: no nlist address for %s\n",
		Pn, Nl[i].n_name);
	    err++;
	  }
	}
	if (err)
	    Exit(1);
/*
 * Open access to kernel memory.
 */
	if (!(Kd = kvm_open(NULL, NULL, NULL, O_RDONLY, Pn)))
	    Exit(1);
/*
 * Drop setgid permission.
 */
	(void) dropgid();
/*
 * Read the inode hash table information.
 */
	if (kread((KA_T)Nl[X_IHEAD].n_value, (char *)&Ihdp, sizeof(Ihdp))) {
	    (void) fprintf(stderr, "%s: can't read *ihead\n", Pn);
	    Exit(1);
	}
	if (!Ihdp) {
	    (void) fprintf(stderr, "%s: *ihead is NUL\n", Pn);
	    Exit(1);
	}
	if (kread((KA_T)Nl[X_INOHSZ].n_value, (char *)&Ihsz, sizeof(Ihsz))) {
	    (void) fprintf(stderr, "%s: can't read inohsz\n", Pn);
	}
}


/*
 * initialize() -- initialize
 */

void
initialize()
{
	(void) get_kernel_access();

#if	solaris>=20500
/*
 * Read the Solaris 2.5 lock manager sysid information.
 */
	if (readsysids())
	    Exit(1);
#endif	/* solaris>=20500 */

}


/*
 * kread() -- read kernel memory
 */

int
kread(addr, buf, len)
	KA_T addr;			/* kernel address */
	char *buf;			/* local receiving buffer address */
	int len;			/* length to read */
{
	int br;

#if	solaris<20500
	if ((unsigned long)addr < (unsigned long)KERNELBASE)
#else	/* solaris>=20500 */
	if ((unsigned long)addr < (unsigned long)Nl[X_KERNELBASE].n_value)
#endif	/* solaris<20500 */

	    return(1);

	br = kvm_read(Kd, (unsigned long)addr, buf, len);
	return((br == len) ? 0 : 1);
}


/*
 * print_dev() -- print device number
 */

char *
print_dev(lp)
	struct llock_info *lp;		/* local lock structure */
{
	static char buf[128];

	(void) sprintf(buf, "%d,%d", (lp->dev >> L_BITSMINOR) & 0x3fff,
	    (lp->dev & L_MAXMIN));
	return(buf);
}

#if	solaris>=20500


/*
 * readsysids() -- read the lock manager's sysid information
 */

static int
readsysids()
{
	char buf[MAXPATHLEN], *cp, *hn, nm[MAXPATHLEN];
	int err;
	KA_T ka;
	MALLOC_S len;
	struct lhaddr *lh;
	int na = 0;
	struct lm_sysid s;
/*
 * Read the lm_sysids pointer from the kernel.  If none is available,
 * make sure the Sids table is empty.
 */
	if (!Nl[X_SYSIDS].n_value) {
	   Nsid = 0;
	   Sids = (struct SysIDs *)NULL;
	   return(0);
	}
	if (kread((KA_T)Nl[X_SYSIDS].n_value, (char *)&ka, sizeof(ka))) {
	    (void) fprintf(stderr, "%s: can't read %s\n", Pn,
		Nl[X_SYSIDS].n_name);
	    return(1);
	}
/*
 * Read the lm_sysid entries linked from the lm_sysids pointer.
 */
	for (; ka; ka = (KA_T)s.next) {
	    if (kread(ka, (char *)&s, sizeof(s)))
		break;
	/*
	 * Read the address buffer.
	 */
	    if (s.addr.len > sizeof(buf) || !s.addr.buf
	    ||  kread((KA_T)s.addr.buf, buf, s.addr.len))
		 buf[0] = '\0';
	/*
	 * Read the name.
	 */
	    if (kread((KA_T)s.name, nm, sizeof(nm)))
		nm[0] = '\0';
	    if (buf[0]) {

	    /*
	     * If the address buffer contains information, truncate
	     * trailing blanks and a last dot.
	     */
		buf[s.addr.len] = '\0';
		for (cp = &buf[s.addr.len] - 1; cp >= buf; cp--) {
		    if (*cp != ' ' && *cp != '.')
			break;
		*cp = '\0';
		}
		cp = buf;
	    } else if (nm[0]) {

	    /*
	     * If the address buffer is empty, use the name buffer if
	     * it's not empty, too.
	     */
		cp = nm;
		nm[sizeof(nm) - 1] = '\0';
	    } else

	    /*
	     * If the address and name buffers are both empty, skip this
	     * lm_sysid entry.
	     */
		continue;
	/*
	 * See if there's room for another local sysid entry.
	 */
	    if (Nsid >= na) {

	    /*
	     * Increase the size of the local sysid table and allocate space
	     * for it.
	     */
		na += 10;
		len = (MALLOC_S)(na * sizeof(struct SysIDs));
		if (!Sids)
		    Sids = (struct SysIDs *)malloc(len);
		else
		    Sids = (struct SysIDs *)realloc(Sids, len);
		if (!Sids) {
		    (void) fprintf(stderr,
			"%s: can't allocate %d SysIDs structures\n", Pn, na);
		    return(1);
		}
	    }
	/*
	 * Allocate space for the address name.
	 */
	    if (!(hn = (char *)malloc(strlen(cp)+1))) {
		(void) fprintf(stderr,
		    "%s: can't allocate space for ID %ld name: %s\n",
		    Pn, s.sysid, cp);
		return(1);
	    }
	    (void) strcpy(hn, cp);
	/*
	 * Look up the host name's address.
	 */
	    if (Oconv) {
		lh = get_haddr(0, hn, (unsigned long)0, &err);
		Sids[Nsid].na = lh->na;
		if (err && Owarn)
		    (void) fprintf(stderr,
			"%s: WARNING: can't get address of %s\n", Pn, hn);
	    }
	/*
	 * Create the local sysid entry.
	 */
	    Sids[Nsid].id = s.sysid;
	    Sids[Nsid++].hn = hn;
	}
	return(0);
}
#endif	/* solaris>=20500 */


/*
 * savelock() -- save lock information
 */

static void
savelock(ip, lp)
	struct inode *ip;		/* inode structure pointer */
	struct lock_descriptor *lp;	/* lock_descriptor structure pointer */
{
	char *cp, *hn, *mn;
	struct hostent *he;
	int i, sel, src;
/*
 * Allocate a local lock structure.
 */
	(void) alloc_llock();
/*
 * Save: inode number; device; size; and process ID.
 */
	Lp->dev = ip->i_dev;
	Lp->inum = ip->i_number;
	Lp->pid = (unsigned long)lp->LOCK_PID;
	Lp->sz = (unsigned long)ip->i_size;
	Lp->szs = 1;
/*
 * Save the lock description: mandatory status; type; start; whence and
 * length or end.
 */
	Lp->mand = MANDLOCK(&ip->i_vnode, ip->i_mode) ? 1 : 0;
	switch (lp->LOCK_TYPE) {
	case F_RDLCK:
	    Lp->type = 1;
	    break;
	case F_WRLCK:
	    Lp->type = 2;
	    break;
	case (F_RDLCK|F_WRLCK):
	    Lp->type = 3;
	    break;
	default:
	    Lp->type = 0;
	}
	Lp->ss = 1;
	Lp->start = (unsigned long)lp->LOCK_START;

#if	defined(LOCK_WHENCE)
	Lp->ls = Lp->ws = 1;
	Lp->len = (unsigned long)lp->LOCK_LEN;
	Lp->whence = (unsigned long)lp->LOCK_WHENCE;
#else	/* !defined(LOCK_WHENCE) */
	Lp->es = 1;
	Lp->end = (unsigned long)lp->LOCK_END;
#endif	/* defined(LOCK_WHENCE) */

#if	solaris<20500
/*
 * If there's a sysid, indicating a remote lock, for Solaris 2.[34], save it.
 */
	if (lp->LOCK_SYSID) {
	    Lp->src = 1;
	    Lp->hn = (char *)NULL;
	    Lp->iap = alloc_in_addr();
	    Lp->iap->s_addr = (unsigned long)lp->LOCK_SYSID;
	}
#else	/* solaris>=20500 */
/*
 * If there's a sysid, indicating a remote lock, for Solaris 2.5 see if
 * there's an entry in the local sysid information table whose sysid
 * matches.  If there is, save the associated host name and address number,
 * otherwise set an "unknown-remote" source type and save the sysid as
 * the address number.
 */
	if (lp->LOCK_SYSID) {
	    Lp->iap = alloc_in_addr();
	    for (i = 0; i < Nsid; i++) {
		if (Sids[i].id == lp->LOCK_SYSID)
		    break;
	    }
	    if (i < Nsid) {
	        Lp->src = 1;
		Lp->hn = Sids[i].hn;
	        Lp->iap->s_addr = (unsigned long)Sids[i].na;
	    } else {
		Lp->src = 2;
		Lp->hn = (char *)NULL;
		Lp->iap->s_addr = lp->LOCK_SYSID;
	    }
	}
#endif	/* solaris<20500 */

	else
	    Lp->src = 0;
}
